@@ -52,6 +52,11 @@
 #define arg5	PLATFORM_SYSCALL_ARG5(tst->arch)
 #define arg6	PLATFORM_SYSCALL_ARG6(tst->arch)
 
+#define VG_ROUNDDN( p, a )   ( ( Addr )( p ) & ~( ( Addr )( a )-1 ) )
+#define VG_ROUNDUP( p, a )   VG_ROUNDDN( ( p )+( a )-1, ( a ) )
+
+#define LOHI64( lo,hi )   (  ( lo ) | ( ( ULong )( hi ) << 32 ) )
+
 #define PRINT(format, args...)  \
    if (VG_(clo_trace_syscalls))        \
       VG_(printf)(format, ## args)
@@ -511,6 +516,86 @@ POST(sys_epoll_wait)
       POST_MEM_WRITE( arg2, sizeof(struct epoll_event)*res ) ;
 }
 
+PRE(sys_fadvise64, 0)
+{
+   PRINT("sys_fadvise64 ( %d, %lld, %lu, %d )",
+         arg1, LOHI64(arg2,arg3), arg4, arg5);
+   PRE_REG_READ5(long, "fadvise64",
+                 int, fd, vki_u32, offset_low, vki_u32, offset_high,
+                 vki_size_t, len, int, advice);
+}
+
+PRE(sys_fadvise64_64, 0)
+{
+   PRINT("sys_fadvise64_64 ( %d, %lld, %lld, %d )",
+         arg1, LOHI64(arg2,arg3), LOHI64(arg4,arg5), arg6);
+   PRE_REG_READ6(long, "fadvise64_64",
+                 int, fd, vki_u32, offset_low, vki_u32, offset_high,
+                 vki_u32, len_low, vki_u32, len_high, int, advice);
+}
+
+PRE(sys_set_mempolicy, 0)
+{
+   PRINT("sys_set_mempolicy ( %d, %p, %d )", arg1,arg2,arg3);
+   PRE_REG_READ3(long, "set_mempolicy",
+                 int, policy, unsigned long *, nodemask,
+                 unsigned long, maxnode);
+   PRE_MEM_READ( "set_mempolicy(nodemask)", arg2,
+                 VG_ROUNDUP( arg3, sizeof(UWord) ) / sizeof(UWord) );
+}
+
+PRE(sys_get_mempolicy, 0)
+{
+   PRINT("sys_get_mempolicy ( %p, %p, %d, %p, %x )", arg1,arg2,arg3,arg4,arg5);
+   PRE_REG_READ5(long, "get_mempolicy",
+                 int *, policy, unsigned long *, nodemask,
+                 unsigned long, maxnode, unsigned long, addr,
+                 unsigned long, flags);
+   if (arg1 != 0)
+      PRE_MEM_WRITE( "get_mempolicy(policy)", arg1, sizeof(Int) );
+   if (arg2 != 0)
+      PRE_MEM_WRITE( "get_mempolicy(nodemask)", arg2,
+                     VG_ROUNDUP( arg3, sizeof(UWord) * 8 ) / sizeof(UWord) );
+}
+
+POST(sys_get_mempolicy)
+{
+   if (arg1 != 0)
+      POST_MEM_WRITE( arg1, sizeof(Int) );
+   if (arg2 != 0)
+      POST_MEM_WRITE( arg2, VG_ROUNDUP( arg3, sizeof(UWord) * 8 ) / sizeof(UWord) );
+}
+
+PRE(sys_inotify_init, 0)
+{
+   PRINT("sys_inotify_init ( )");
+   PRE_REG_READ0(long, "inotify_init");
+}
+
+POST(sys_inotify_init)
+{
+   if (!VG_(fd_allowed)(res, "inotify_init", tid, True)) {
+      VG_(close)(res);
+      set_result( -VKI_EMFILE );
+   } else {
+      if (VG_(clo_track_fds))
+         VG_(record_fd_open) (tid, res, NULL);
+   }
+}
+
+PRE(sys_inotify_add_watch, 0)
+{
+   PRINT( "sys_inotify_add_watch ( %d, %p, %x )", arg1,arg2,arg3);
+   PRE_REG_READ3(long, "inotify_add_watch", int, fd, char *, path, int, mask);
+   PRE_MEM_RASCIIZ( "inotify_add_watch(path)", arg2 );
+}
+
+PRE(sys_inotify_rm_watch, 0)
+{
+   PRINT( "sys_inotify_rm_watch ( %d, %x )", arg1,arg2);
+   PRE_REG_READ2(long, "inotify_rm_watch", int, fd, int, wd);
+}
+
 // Nb: this wrapper is "Special" because we have to pad/unpad memory around
 // the syscall itself, and this allows us to control exactly the code that
 // gets run while the padding is in place.
