@@ -1,13 +1,14 @@
 
 /*--------------------------------------------------------------------*/
-/*--- User-mode execve()                                     ume.c ---*/
+/*--- User-mode execve(), and other stuff shared between stage1    ---*/
+/*--- and stage2.                                            ume.c ---*/
 /*--------------------------------------------------------------------*/
 
 /*
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -32,23 +33,26 @@
 #define _GNU_SOURCE
 #define _FILE_OFFSET_BITS 64
 
-#include "vg_include.h"
+#include "core.h"
 
-#include <stddef.h>
 #include <sys/mman.h>
 #include <fcntl.h>
 #include <errno.h>
-#include <elf.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <sys/stat.h>
-#include <dlfcn.h>
 #include <assert.h>
 
 #include "ume.h"
-#include "vg_include.h"
+
+#if	ELFSZ == 64
+#define ESZ(x)	Elf64_##x
+#elif	ELFSZ == 32
+#define ESZ(x)	Elf32_##x
+#else
+#error ELFSZ needs to ==32 or ==64
+#endif
 
 struct elfinfo
 {
@@ -60,9 +64,7 @@ struct elfinfo
 static void check_mmap(void* res, void* base, int len)
 {
    if ((void*)-1 == res) {
-      fprintf(stderr, "valgrind: mmap(%p, %d) failed during startup.\n"  
-                      "valgrind: is there a hard virtual memory limit set?\n",
-                      base, len);
+      fprintf(stderr, "valgrind: mmap(%p, %d) failed in UME.\n", base, len);
       exit(1);
    }
 }
@@ -106,7 +108,7 @@ void foreach_map(int (*fn)(char *start, char *end,
       int ino;
       void *segstart, *segend;
 
-      sscanf(bufptr, "%p-%p %s %Lx %x:%x %d",
+      sscanf(bufptr, "%p-%p %s %llx %x:%x %d",
 	     &segstart, &segend, perm, &offset, &maj, &min, &ino);
       bufptr = strchr(bufptr, '\n');
       if (bufptr != NULL)
@@ -117,147 +119,23 @@ void foreach_map(int (*fn)(char *start, char *end,
    }
 }
 
-typedef struct {
-   char* fillgap_start;
-   char* fillgap_end;
-   int   fillgap_padfile;
-} fillgap_extra;
-
-static int fillgap(char *segstart, char *segend, const char *perm, off_t off, 
-                   int maj, int min, int ino, void* e)
-{
-   fillgap_extra* extra = e;
-
-   if (segstart >= extra->fillgap_end)
-      return 0;
-
-   if (segstart > extra->fillgap_start) {
-      void* res = mmap(extra->fillgap_start, segstart - extra->fillgap_start,
-                       PROT_NONE, MAP_FIXED|MAP_PRIVATE, 
-                       extra->fillgap_padfile, 0);
-      check_mmap(res, extra->fillgap_start, segstart - extra->fillgap_start);
-   }
-   extra->fillgap_start = segend;
-   
-   return 1;
-}
-
-// Choose a name for the padfile, open it.
-int as_openpadfile(void)
-{
-   char buf[256];
-   int padfile;
-   int seq = 1;
-   do {
-      snprintf(buf, 256, "/tmp/.pad.%d.%d", getpid(), seq++);
-      padfile = open(buf, O_RDWR|O_CREAT|O_EXCL, 0);
-      unlink(buf);
-      if (padfile == -1 && errno != EEXIST) {
-         fprintf(stderr, "valgrind: couldn't open padfile\n");
-         exit(44);
-      }
-   } while(padfile == -1);
-
-   return padfile;
-}
-
-// Pad all the empty spaces in a range of address space to stop interlopers.
-void as_pad(void *start, void *end, int padfile)
-{
-   fillgap_extra extra;
-   extra.fillgap_start   = start;
-   extra.fillgap_end     = end;
-   extra.fillgap_padfile = padfile;
-
-   foreach_map(fillgap, &extra);
-	
-   if (extra.fillgap_start < extra.fillgap_end) {
-      void* res = mmap(extra.fillgap_start, 
-                       extra.fillgap_end - extra.fillgap_start,
-                       PROT_NONE, MAP_FIXED|MAP_PRIVATE, padfile, 0);
-      check_mmap(res, extra.fillgap_start, 
-                 extra.fillgap_end - extra.fillgap_start);
-   }
-}
-
-typedef struct {
-   char*        killpad_start;
-   char*        killpad_end;
-   struct stat* killpad_padstat;
-} killpad_extra;
-
-static int killpad(char *segstart, char *segend, const char *perm, off_t off, 
-                   int maj, int min, int ino, void* ex)
-{
-   killpad_extra* extra = ex;
-   void *b, *e;
-   int res;
-
-   assert(NULL != extra->killpad_padstat);
-
-   if (extra->killpad_padstat->st_dev != makedev(maj, min) || 
-       extra->killpad_padstat->st_ino != ino)
-      return 1;
-   
-   if (segend <= extra->killpad_start || segstart >= extra->killpad_end)
-      return 1;
-   
-   if (segstart <= extra->killpad_start)
-      b = extra->killpad_start;
-   else
-      b = segstart;
-   
-   if (segend >= extra->killpad_end)
-      e = extra->killpad_end;
-   else
-      e = segend;
-   
-   res = munmap(b, (char *)e-(char *)b);
-   assert(0 == res);
-   
-   return 1;
-}
-
-// Remove padding of 'padfile' from a range of address space.
-void as_unpad(void *start, void *end, int padfile)
-{
-   static struct stat padstat;
-   killpad_extra extra;
-   int res;
-
-   assert(padfile > 0);
-   
-   res = fstat(padfile, &padstat);
-   assert(0 == res);
-   extra.killpad_padstat = &padstat;
-   extra.killpad_start   = start;
-   extra.killpad_end     = end;
-   foreach_map(killpad, &extra);
-}
-
-void as_closepadfile(int padfile)
-{
-   int res = close(padfile);
-   assert(0 == res);
-}
-
 /*------------------------------------------------------------*/
 /*--- Finding auxv on the stack                            ---*/
 /*------------------------------------------------------------*/
 
-struct ume_auxv *find_auxv(int *esp)
+struct ume_auxv *find_auxv(UWord* sp)
 {
-   esp++;			/* skip argc */
+   sp++;                // skip argc (Nb: is word-sized, not int-sized!)
 
-   while(*esp != 0)		/* skip argv */
-      esp++;
-   esp++;
+   while (*sp != 0)     // skip argv
+      sp++;
+   sp++;
 
-   while(*esp != 0)		/* skip env */
-      esp++;
-   esp++;
+   while (*sp != 0)     // skip env
+      sp++;
+   sp++;
    
-   return (struct ume_auxv *)esp;
+   return (struct ume_auxv *)sp;
 }
 
 /*------------------------------------------------------------*/
@@ -282,12 +160,13 @@ struct elfinfo *readelf(int fd, const char *filename)
       fprintf(stderr, "valgrind: %s: bad ELF magic\n", filename);
       return NULL;
    }
-   if (e->e.e_ident[EI_CLASS] != ELFCLASS32) {
-      fprintf(stderr, "valgrind: Can only handle 32-bit executables\n");
+   if (e->e.e_ident[EI_CLASS] != VG_ELF_CLASS) {
+      fprintf(stderr, "valgrind: wrong executable class (eg. 32-bit instead\n"
+                      "valgrind: of 64-bit)\n");
       return NULL;
    }
-   if (e->e.e_ident[EI_DATA] != ELFDATA2LSB) {
-      fprintf(stderr, "valgrind: Expecting little-endian\n");
+   if (e->e.e_ident[EI_DATA] != VG_ELF_ENDIANNESS) {
+      fprintf(stderr, "valgrind: wrong endian-ness\n");
       return NULL;
    }
    if (!(e->e.e_type == ET_EXEC || e->e.e_type == ET_DYN)) {
@@ -295,8 +174,8 @@ struct elfinfo *readelf(int fd, const char *filename)
       return NULL;
    }
 
-   if (e->e.e_machine != EM_386) {
-      fprintf(stderr, "valgrind: need x86\n");
+   if (e->e.e_machine != VG_ELF_MACHINE) {
+      fprintf(stderr, "valgrind: wrong architecture\n");
       return NULL;
    }
 
@@ -332,8 +211,8 @@ ESZ(Addr) mapelf(struct elfinfo *e, ESZ(Addr) base)
       if (ph->p_type != PT_LOAD)
 	 continue;
 
-      addr = ph->p_vaddr+base;
-      memsz = ph->p_memsz;
+      addr    = ph->p_vaddr+base;
+      memsz   = ph->p_memsz;
       brkaddr = addr+memsz;
 
       if (brkaddr > elfbrk)
@@ -352,27 +231,25 @@ ESZ(Addr) mapelf(struct elfinfo *e, ESZ(Addr) base)
       if (ph->p_type != PT_LOAD)
 	 continue;
 
-      if (ph->p_flags & PF_X)
-	 prot |= PROT_EXEC;
-      if (ph->p_flags & PF_W)
-	 prot |= PROT_WRITE;
-      if (ph->p_flags & PF_R)
-	 prot |= PROT_READ;
+      if (ph->p_flags & PF_X) prot |= PROT_EXEC;
+      if (ph->p_flags & PF_W) prot |= PROT_WRITE;
+      if (ph->p_flags & PF_R) prot |= PROT_READ;
 
-      align = ph->p_align;
-
-      addr = ph->p_vaddr+base;
-      off =  ph->p_offset;
-      filesz = ph->p_filesz;
-      bss = addr+filesz;
-      memsz = ph->p_memsz;
+      align   = ph->p_align;
+      addr    = ph->p_vaddr+base;
+      off     = ph->p_offset;
+      filesz  = ph->p_filesz;
+      bss     = addr+filesz;
+      memsz   = ph->p_memsz;
       brkaddr = addr+memsz;
-
-      res = mmap((char *)ROUNDDN(addr, align),
-                 ROUNDUP(bss, align)-ROUNDDN(addr, align),
-                 prot, MAP_FIXED|MAP_PRIVATE, e->fd, ROUNDDN(off, align));
-      check_mmap(res, (char*)ROUNDDN(addr,align),
-                 ROUNDUP(bss, align)-ROUNDDN(addr, align));
+      
+      if (ROUNDUP(bss, align)-ROUNDDN(addr, align) > 0) {
+	 res = mmap((char *)ROUNDDN(addr, align),
+		    ROUNDUP(bss, align)-ROUNDDN(addr, align),
+		    prot, MAP_FIXED|MAP_PRIVATE, e->fd, ROUNDDN(off, align));
+	 check_mmap(res, (char*)ROUNDDN(addr,align),
+		    ROUNDUP(bss, align)-ROUNDDN(addr, align));
+      }
 
       /* if memsz > filesz, then we need to fill the remainder with zeroed pages */
       if (memsz > filesz) {
@@ -385,9 +262,9 @@ ESZ(Addr) mapelf(struct elfinfo *e, ESZ(Addr) base)
             check_mmap(res, (char*)ROUNDUP(bss,align), bytes);
          }
 
-	 bytes = bss & (VKI_BYTES_PER_PAGE - 1);
-	 if (bytes > 0) {
-	    bytes = VKI_BYTES_PER_PAGE - bytes;
+	 bytes = bss & (VKI_PAGE_SIZE - 1);
+	 if ((prot & PROT_WRITE) && (bytes > 0)) {
+	    bytes = VKI_PAGE_SIZE - bytes;
 	    memset((char *)bss, 0, bytes);
 	 }
       }
@@ -410,12 +287,18 @@ static int load_ELF(char *hdr, int len, int fd, const char *name,
 {
    struct elfinfo *e;
    struct elfinfo *interp = NULL;
-   ESZ(Addr) minaddr = ~0;
-   ESZ(Addr) maxaddr = 0;
-   ESZ(Addr) interp_addr = 0;
-   ESZ(Word) interp_size = 0;
+   ESZ(Addr) minaddr = ~0;	/* lowest mapped address */
+   ESZ(Addr) maxaddr = 0;	/* highest mapped address */
+   ESZ(Addr) interp_addr = 0;	/* interpreter (ld.so) address */
+   ESZ(Word) interp_size = 0;	/* interpreter size */
+   ESZ(Word) interp_align = VKI_PAGE_SIZE;
    int i;
    void *entry;
+   ESZ(Addr) ebase = 0;
+
+#ifdef HAVE_PIE
+   ebase = info->exe_base;
+#endif
 
    e = readelf(fd, name);
 
@@ -423,14 +306,15 @@ static int load_ELF(char *hdr, int len, int fd, const char *name,
       return ENOEXEC;
 
    info->phnum = e->e.e_phnum;
-   info->entry = e->e.e_entry;
+   info->entry = e->e.e_entry + ebase;
+   info->phdr = 0;
 
    for(i = 0; i < e->e.e_phnum; i++) {
       ESZ(Phdr) *ph = &e->p[i];
 
       switch(ph->p_type) {
       case PT_PHDR:
-	 info->phdr = ph->p_vaddr;
+	 info->phdr = ph->p_vaddr + ebase;
 	 break;
 
       case PT_LOAD:
@@ -472,7 +356,8 @@ static int load_ELF(char *hdr, int len, int fd, const char *name,
 	       continue;
 	    
 	    if (!baseaddr_set) {
-	       interp_addr = iph->p_vaddr;
+	       interp_addr  = iph->p_vaddr;
+	       interp_align = iph->p_align;
 	       baseaddr_set = 1;
 	    }
 
@@ -483,23 +368,30 @@ static int load_ELF(char *hdr, int len, int fd, const char *name,
 	       interp_size = end;
 	 }
 	 break;
+
+      default:
+         // do nothing
+         break;
       }
       }
    }
 
+   if (info->phdr == 0)
+      info->phdr = minaddr + e->e.e_phoff;
+
    if (info->exe_base != info->exe_end) {
       if (minaddr >= maxaddr ||
-	  (minaddr < info->exe_base ||
-	   maxaddr > info->exe_end)) {
+	  (minaddr + ebase < info->exe_base ||
+	   maxaddr + ebase > info->exe_end)) {
 	 fprintf(stderr, "Executable range %p-%p is outside the\n"
                          "acceptable range %p-%p\n",
-		 (void *)minaddr,        (void *)maxaddr,
-		 (void *)info->exe_base, (void *)info->exe_end);
+		 (void *)minaddr + ebase, (void *)maxaddr + ebase,
+		 (void *)info->exe_base,  (void *)info->exe_end);
 	 return ENOMEM;
       }
    }
 
-   info->brkbase = mapelf(e, 0);		/* map the executable */
+   info->brkbase = mapelf(e, ebase);	/* map the executable */
 
    if (info->brkbase == 0)
       return ENOMEM;
@@ -512,7 +404,7 @@ static int load_ELF(char *hdr, int len, int fd, const char *name,
       int flags = MAP_PRIVATE|MAP_ANONYMOUS;
 
       if (info->map_base != 0) {
-	 base = (char *)info->map_base;
+	 base = (char *)ROUNDUP(info->map_base, interp_align);
 	 flags |= MAP_FIXED;
       }
 
@@ -525,18 +417,21 @@ static int load_ELF(char *hdr, int len, int fd, const char *name,
       mapelf(interp, (ESZ(Addr))baseoff);
 
       close(interp->fd);
-      free(interp);
 
       entry = baseoff + interp->e.e_entry;
       info->interp_base = (ESZ(Addr))base;
+
+      free(interp->p);
+      free(interp);
    } else
       entry = (void *)e->e.e_entry;
 
-   info->exe_base = minaddr;
-   info->exe_end = maxaddr;
+   info->exe_base = minaddr + ebase;
+   info->exe_end  = maxaddr + ebase;
 
-   info->init_eip = (addr_t)entry;
+   info->init_eip = (Addr)entry;
 
+   free(e->p);
    free(e);
 
    return 0;
@@ -601,14 +496,67 @@ static int load_script(char *hdr, int len, int fd, const char *name,
    return do_exec_inner(interp, info);
 }
 
+/* 
+   Emulate the normal Unix permissions checking algorithm.
+
+   If owner matches, then use the owner permissions, else
+   if group matches, then use the group permissions, else
+   use other permissions.
+
+   Note that we can't deal with SUID/SGID, so we refuse to run them
+   (otherwise the executable may misbehave if it doesn't have the
+   permissions it thinks it does).
+*/
+static int check_perms(int fd)
+{
+   struct stat st;
+
+   if (fstat(fd, &st) == -1) 
+      return errno;
+
+   if (st.st_mode & (S_ISUID | S_ISGID)) {
+      //fprintf(stderr, "Can't execute suid/sgid executable %s\n", exe);
+      return EACCES;
+   }
+
+   if (geteuid() == st.st_uid) {
+      if (!(st.st_mode & S_IXUSR))
+	 return EACCES;
+   } else {
+      int grpmatch = 0;
+
+      if (getegid() == st.st_gid)
+	 grpmatch = 1;
+      else {
+	 gid_t groups[32];
+	 int ngrp = getgroups(32, groups);
+	 int i;
+
+	 for(i = 0; i < ngrp; i++)
+	    if (groups[i] == st.st_gid) {
+	       grpmatch = 1;
+	       break;
+	    }
+      }
+
+      if (grpmatch) {
+	 if (!(st.st_mode & S_IXGRP))
+	    return EACCES;
+      } else if (!(st.st_mode & S_IXOTH))
+	 return EACCES;
+   }
+
+   return 0;
+}
+
 static int do_exec_inner(const char *exe, struct exeinfo *info)
 {
    int fd;
-   char buf[VKI_BYTES_PER_PAGE];
+   int err;
+   char buf[VKI_PAGE_SIZE];
    int bufsz;
    int i;
    int ret;
-   struct stat st;
    static const struct {
       int (*match)(const char *hdr, int len);
       int (*load) (      char *hdr, int len, int fd2, const char *name,
@@ -626,40 +574,10 @@ static int do_exec_inner(const char *exe, struct exeinfo *info)
       return errno;
    }
 
-   if (fstat(fd, &st) == -1) 
-      return errno;
-   else {
-      uid_t uid = geteuid();
-      gid_t gid = getegid();
-      gid_t groups[32];
-      int ngrp = getgroups(32, groups);
-
-      if (st.st_mode & (S_ISUID | S_ISGID)) {
-	 fprintf(stderr, "Can't execute suid/sgid executable %s\n", exe);
-	 return EACCES;
-      }
-
-      if (uid == st.st_uid) {
-	 if (!(st.st_mode & S_IXUSR))
-	    return EACCES;
-      } else {
-	 int grpmatch = 0;
-
-	 if (gid == st.st_gid)
-	    grpmatch = 1;
-	 else 
-	    for(i = 0; i < ngrp; i++)
-	       if (groups[i] == st.st_gid) {
-		  grpmatch = 1;
-		  break;
-	       }
-
-	 if (grpmatch) {
-	    if (!(st.st_mode & S_IXGRP))
-	       return EACCES;
-	 } else if (!(st.st_mode & S_IXOTH))
-	    return EACCES;
-      }
+   err = check_perms(fd);
+   if (err != 0) {
+      close(fd);
+      return err;
    }
 
    bufsz = pread(fd, buf, sizeof(buf), 0);
