@@ -9,7 +9,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Nicholas Nethercote
+   Copyright (C) 2000-2005 Nicholas Nethercote
       njn25@cam.ac.uk
 
    This program is free software; you can redistribute it and/or
@@ -31,7 +31,7 @@
 */
 
 
-#include "vg_include.h"
+#include "core.h"
 
 /* ---------------------------------------------------------------------
    Error messages (for malformed tools)
@@ -66,9 +66,9 @@ void malloc_panic ( const Char* fn )
 /*--- Replacing malloc et al                               ---*/
 /*------------------------------------------------------------*/
 
-/* Default redzone for CLIENT arena of Valgrind's malloc() is 4 bytes */
+/* Default redzone size for CLIENT arena of Valgrind's malloc() */
 __attribute__ ((weak))
-UInt VG_(vg_malloc_redzone_szB) = 4;
+UInt VG_(vg_malloc_redzone_szB) = 8;
 
 Bool VG_(sk_malloc_called_by_scheduler) = False;
 
@@ -78,17 +78,15 @@ Bool VG_(sk_malloc_called_by_scheduler) = False;
    scheduler is the only place this can be called from;  this ensures that a
    malloc()-replacing tool cannot forget to implement SK_(malloc)() or
    SK_(free)().  */
-__attribute__ ((weak))
-void* SK_(malloc)( Int size )
+void* VG_(missing_malloc)( SizeT size )
 {
    if (VG_(sk_malloc_called_by_scheduler))
-      return VG_(cli_malloc)(4, size);
+      return VG_(cli_malloc)(VG_MIN_MALLOC_SZB, size);
    else 
       malloc_panic(__PRETTY_FUNCTION__);
 }
 
-__attribute__ ((weak))
-void  SK_(free)( void* p )
+void VG_(missing_free)( void* p )
 {
    /* see comment for SK_(malloc)() above */
    if (VG_(sk_malloc_called_by_scheduler))
