@@ -8,7 +8,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -30,15 +30,7 @@
 */
 
 
-#include "vg_include.h"
-
-
-/* static ... to keep it out of the stack frame. */
-static Char procmap_buf[M_PROCMAP_BUF];
-
-/* Records length of /proc/self/maps read into procmap_buf. */
-static Int  buf_n_tot;
-
+#include "core.h"
 
 /* Helper fns. */
 
@@ -56,14 +48,14 @@ static Int decdigit ( Char c )
    return -1;
 }
 
-static Int readchar ( Char* buf, Char* ch )
+static Int readchar ( const Char* buf, Char* ch )
 {
    if (*buf == 0) return 0;
    *ch = *buf;
    return 1;
 }
 
-static Int readhex ( Char* buf, UInt* val )
+static Int readhex ( const Char* buf, UWord* val )
 {
    Int n = 0;
    *val = 0;
@@ -74,7 +66,7 @@ static Int readhex ( Char* buf, UInt* val )
    return n;
 }
 
-static Int readdec ( Char* buf, UInt* val )
+static Int readdec ( const Char* buf, UInt* val )
 {
    Int n = 0;
    *val = 0;
@@ -86,10 +78,15 @@ static Int readdec ( Char* buf, UInt* val )
 }
 
 
-/* Read /proc/self/maps, store the contents in a static buffer.  If there's
-   a syntax error or other failure, just abort. */
-void VG_(read_procselfmaps)(void)
+/* Read /proc/self/maps, store the contents in a buffer.  If there's a
+   syntax error or other failure, just abort.  The buffer size is
+   increased dynamically to fit whatever sized contents we need to
+   read.  It can get pretty big in some programs. */
+static const Char *read_procselfmaps(Int *maps_size)
 {
+   static Char *buffer;
+   static Int bufsize = 1024;
+   Int bufused = 0;
    Int n_chunk, fd;
    
    /* Read the initial memory mapping from the /proc filesystem. */
@@ -98,23 +95,36 @@ void VG_(read_procselfmaps)(void)
       VG_(message)(Vg_UserMsg, "FATAL: can't open /proc/self/maps");
       VG_(exit)(1);
    }
-   buf_n_tot = 0;
+
+   bufused = 0;
+   if (buffer == NULL)
+      buffer = VG_(arena_malloc)(VG_AR_CORE, bufsize);
+
    do {
-      n_chunk = VG_(read) ( fd, &procmap_buf[buf_n_tot], 
-                            M_PROCMAP_BUF - buf_n_tot );
-      buf_n_tot += n_chunk;
-   } while ( n_chunk > 0 && buf_n_tot < M_PROCMAP_BUF );
+      Int want = bufsize - bufused;
+      n_chunk = VG_(read) ( fd, &buffer[bufused], want );
+
+      if (n_chunk < 0) {
+	 VG_(message)(Vg_UserMsg, "FATAL: read error on /proc/self/maps");
+	 VG_(exit)(99);
+      }
+
+      bufused += n_chunk;
+
+      if (bufused == bufsize) {
+	 bufsize *= 2;
+	 buffer = VG_(arena_realloc)(VG_AR_CORE, buffer, VG_MIN_MALLOC_SZB, bufsize);
+      }
+   } while ( n_chunk > 0 );
    VG_(close)(fd);
-   if (buf_n_tot >= M_PROCMAP_BUF-5) {
-      VG_(message)(Vg_UserMsg, "FATAL: M_PROCMAP_BUF is too small; "
-                               "increase it and recompile");
-       VG_(exit)(1);
-   }
-   if (buf_n_tot == 0) {
+   if (bufused == 0) {
       VG_(message)(Vg_UserMsg, "FATAL: I/O error on /proc/self/maps" );
        VG_(exit)(1);
    }
-   procmap_buf[buf_n_tot] = 0;
+   buffer[bufused] = '\0';
+
+   *maps_size = bufused;
+   return buffer;
 }
 
 /* Parse /proc/self/maps.  For each map entry, call
@@ -122,35 +132,36 @@ void VG_(read_procselfmaps)(void)
 
       start address in memory
       length
-      r permissions char; either - or r
-      w permissions char; either - or w
-      x permissions char; either - or x
+      page protections (using the VKI_PROT_* flags)
+      mapped file device and inode
       offset in file, or zero if no file
       filename, zero terminated, or NULL if no file
 
    So the sig of the called fn might be
 
-      void (*record_mapping)( Addr start, UInt size, 
-                              Char r, Char w, Char x, 
-                              UInt foffset, UChar* filename )
+      void (*record_mapping)( Addr start, SizeT size, UInt prot,
+			      UInt dev, UInt info,
+                              ULong foffset, UChar* filename )
 
    Note that the supplied filename is transiently stored; record_mapping 
    should make a copy if it wants to keep it.
-
-   Nb: it is important that this function does not alter the contents of
-       procmap_buf!
 */
 void VG_(parse_procselfmaps) (
-   void (*record_mapping)( Addr addr, UInt len, Char rr, Char ww, Char xx, 
+   void (*record_mapping)( Addr addr, SizeT len, UInt prot,
 			   UInt dev, UInt ino, ULong foff, const UChar* filename )
    )
 {
    Int    i, j, i_eol;
    Addr   start, endPlusOne;
    UChar* filename;
-   UInt   foffset;
    UChar  rr, ww, xx, pp, ch, tmp;
-   UInt	  maj, min, ino;
+   UInt	  ino;
+   UWord  foffset, maj, min;
+   const Char *procmap_buf;
+   Int    buf_n_tot;
+   UInt   prot;
+
+   procmap_buf = read_procselfmaps(&buf_n_tot);
 
    sk_assert( '\0' != procmap_buf[0] && 0 != buf_n_tot);
 
@@ -220,31 +231,41 @@ void VG_(parse_procselfmaps) (
     read_line_ok:
 
       /* Try and find the name of the file mapped to this segment, if
-         it exists. */
-      while (procmap_buf[i] != '\n' && i < M_PROCMAP_BUF-1) i++;
+         it exists.  Note that files can contains spaces. */
+
+      // Move i to the next non-space char, which should be either a '/' or
+      // a newline.
+      while (procmap_buf[i] == ' ' && i < buf_n_tot-1) i++;
+     
+      // Move i_eol to the end of the line.
       i_eol = i;
-      i--;
-      while (!VG_(isspace)(procmap_buf[i]) && i >= 0) i--;
-      i++;
+      while (procmap_buf[i_eol] != '\n' && i_eol < buf_n_tot-1) i_eol++;
+
+      // If there's a filename...
       if (i < i_eol-1 && procmap_buf[i] == '/') {
          /* Minor hack: put a '\0' at the filename end for the call to
             `record_mapping', then restore the old char with `tmp'. */
-         filename = &procmap_buf[i];
-         tmp = filename[i_eol - i];
+	 filename = VG_(arena_malloc)(VG_AR_CORE, i_eol-i+1);
+	 VG_(memcpy)(filename, &procmap_buf[i], i_eol-i);
          filename[i_eol - i] = '\0';
       } else {
-         tmp = '\0';
+	 tmp = 0;
          filename = NULL;
          foffset = 0;
       }
 
-      (*record_mapping) ( start, endPlusOne-start, 
-                          rr, ww, xx, maj * 256 + min, ino,
-                          foffset, filename );
+      prot = 0;
+      if (rr == 'r') prot |= VKI_PROT_READ;
+      if (ww == 'w') prot |= VKI_PROT_WRITE;
+      if (xx == 'x') prot |= VKI_PROT_EXEC;
 
-      if ('\0' != tmp) {
-         filename[i_eol - i] = tmp;
-      }
+      if (start < VG_(valgrind_last))
+	 (*record_mapping) ( start, endPlusOne-start, 
+			     prot, maj * 256 + min, ino,
+			     foffset, filename );
+
+      if (filename != NULL)
+	 VG_(arena_free)(VG_AR_CORE, filename);
 
       i = i_eol + 1;
    }
