@@ -1,12 +1,12 @@
 /*--------------------------------------------------------------------*/
-/*--- Read stabs debug info.                     vg_symtab_stabs.c ---*/
+/*--- Read stabs debug info.                            vg_stabs.c ---*/
 /*--------------------------------------------------------------------*/
 
 /*
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward
+   Copyright (C) 2000-2004 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -27,7 +27,7 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "vg_include.h"
+#include "core.h"
 #include "vg_symtab2.h"
 
 #include <a.out.h>        /* stabs defns                    */
@@ -101,6 +101,7 @@ typedef struct _StabTypeTab {
    }		*headerhash[HEADER_HASHSZ];
 } StabTypeTab;
 
+static const Bool stabs_debug = False;
 
 static UInt header_hash(Char *filename, UInt instance)
 {
@@ -122,7 +123,7 @@ static UInt header_hash(Char *filename, UInt instance)
 */
 static SymType *structRef(StabTypeTab *tab, SymType *def, Bool isstruct, Char *name)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    struct structlist *sl;
    SymType *ty;
    static Int warnlen = 0;
@@ -167,14 +168,24 @@ static SymType *structRef(StabTypeTab *tab, SymType *def, Bool isstruct, Char *n
 /* Add a structural defintion for a struct/union reference */
 static SymType *structDef(StabTypeTab *tab, SymType *def, Bool isstruct, Char *name)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    SymType *ref = structRef(tab, NULL, isstruct, name);
 
-   if (debug)
-      VG_(printf)("defining %s ref for %s %p -> %p\n",
-		  isstruct ? "struct" : "union", name, ref, def);
+   /* it seems that GNAT likes to declare names as both struct tags
+      and typedefs so check we aren't about to make a structure a
+      reference to itself as that will create a loop */
+   if (ref == def) {
+      if (debug)
+         VG_(printf)("ignoring %s self ref for %s %p -> %p\n",
+		     isstruct ? "struct" : "union", name, ref, def);
+   }
+   else {
+      if (debug)
+         VG_(printf)("defining %s ref for %s %p -> %p\n",
+		     isstruct ? "struct" : "union", name, ref, def);
 
-   def = VG_(st_mktypedef)(ref, name, VG_(st_basetype)(def, False));
+      def = VG_(st_mktypedef)(ref, name, VG_(st_basetype)(def, False));
+   }
    VG_(st_setname)(def, name);
    return def;
 }
@@ -225,7 +236,7 @@ static StabFile *getStabFile(StabTypeTab *tab, Int file, StabFile *set)
 /* add a new index for a file */
 static void addFileAlias(StabTypeTab *tab, Char *filename, UInt instance, Int idx)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    struct header *hp;
 
    for(hp = tab->headerhash[header_hash(filename, instance)]; hp != NULL; hp = hp->next) {
@@ -244,7 +255,7 @@ static void addFileAlias(StabTypeTab *tab, Char *filename, UInt instance, Int id
 
 static void addHeader(StabTypeTab *tab, Char *filename, UInt instance, Int idx)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    struct header *hp, **bucket;
    
    if (debug)
@@ -300,35 +311,41 @@ static StabType *getStabType(StabTypeTab *tab, Int file, Int sym)
    return &sf->types[sym];
 }
 
-static inline Bool isdigit(Char c, Int base, Int *v)
+static Bool isdigit_base(Char c, Int base, Int *vp)
 {
+   Bool ret = False;
+   Int v = 0;
+
    switch(base) {
    case 10:
    case 0:
-      *v = c - '0';
-      return c >= '0' && c <= '9';
+      v = c - '0';
+      ret = (c >= '0' && c <= '9');
+      break;
 
    case 8:
-      *v = c - '0';
-      return c >= '0' && c <= '7';
+      v = c - '0';
+      ret = (c >= '0' && c <= '7');
+      break;
 
    case 16:
       if (c >= '0' && c <= '9') {
-	 *v = c - '0';
-	 return True;
+	 v = c - '0';
+	 ret = True;
+      } else if (c >= 'a' && c <= 'f') {
+	 v = c - 'a';
+	 ret = True;
+      } else if (c >= 'A' && c <= 'F') {
+	 v = c - 'F';
+	 ret = True;
       }
-      if (c >= 'a' && c <= 'f') {
-	 *v = c - 'a';
-	 return True;
-      }
-      if (c >= 'A' && c <= 'F') {
-	 *v = c - 'F';
-	 return True;
-      }
-      return False;
+      break;
    }
 
-   return False;
+   if (vp && ret)
+      *vp = v;
+
+   return ret;
 }
 
 static inline Int getbase(Char **pp)
@@ -365,7 +382,7 @@ static Int atoi(Char **pp, Int base)
    if (base == 0)
       base = getbase(&p);
 
-   while(isdigit(*p, base, &v)) {
+   while(isdigit_base(*p, base, &v)) {
       ret *= base;
       ret += v;
       p++;
@@ -386,7 +403,7 @@ static UInt atou(Char **pp, Int base)
    if (base == 0)
       base = getbase(&p);
 
-   while(isdigit(*p, base, &v)) {
+   while(isdigit_base(*p, base, &v)) {
       ret *= base;
       ret += v;
       p++;
@@ -396,12 +413,36 @@ static UInt atou(Char **pp, Int base)
    return ret;
 }
 
-/* Skip to end of name (':'); "::" in the name indicate nesting, and
-   are included within the name */
-static Char *nested_name(Char *p) {
-   while(*p && (p[0] != ':' || p[1] == ':')) {
-      if (*p == ':')
-	 p++;
+/* Skip a ':'-delimited name which may have ::, 'char' or other things in
+   <> brackets */
+static Char *templ_name(Char *p)
+{
+   Int brac = 0;
+
+   for(;;) {
+      if (*p == '<')
+	 brac++;
+      if (*p == '>')
+	 brac--;
+      /* skip quoted character (note, it could be anything, even a
+	 literal \0)
+
+	 XXX This is a complete botch; we can't do anything sane here,
+	 like support \-quoting, because gcc doesn't seem to generate
+	 it, and even if it did, we wouldn't know what "'\'" means -
+	 the begining of '\'' or a char in itself ('\\')?
+       */
+      if (brac && p[0] == '\'' && p[2] == '\'')
+	 p += 3;
+
+      /* If we're within <>, then treat :: as part of the name (a single
+	 : still terminates) */
+      if (*p == ':') {
+	 if (brac && p[1] == ':' && p[-1] != '<')
+	    p++;
+	 else
+	    break;
+      }
       p++;
    }
 
@@ -435,7 +476,7 @@ static void parse_typeref(Char **pp, Int *filep, Int *symp)
 
 static void stab_resolve(SymType *st, void *data)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    Char *str = (Char *)data;
    vg_assert(!VG_(st_isresolved)(st));
 
@@ -452,7 +493,7 @@ static void stab_resolve(SymType *st, void *data)
    introduced anywhere, so we need to scan it all to pick them up. */
 static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    Char *p = *pp;
    Char t;
    SymType *type;
@@ -534,8 +575,11 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
 	 if (debug)
 	    VG_(printf)("defining type %p (%d,%d) = %s\n", symtype, file, sym, p);
 
-	 /* skip type attributes */
-	 while(*p == '@')
+	 /* Skip type attributes
+	    '@' could also be pointer-to-member, so we need to see if
+	    the following character looks like a type reference or not.
+	  */
+	 while(*p == '@' && !(VG_(isdigit)(p[1]) || p[1] == '-' || p[1] == '(') )
 	    p = SKIPPAST(p+1, ';', "type attrib");
 
 	 prev = p;
@@ -577,14 +621,48 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
       p--;
       n = atoi(&p, 0);
       switch(n) {
-      case -16:	type = VG_(st_mkbool)(def, sizeof(int)); break;
+      case -1:	type = VG_(st_mkint)(def, 4, True); break;
+      case -2:	type = VG_(st_mkint)(def, 1, True); break;
+      case -3:	type = VG_(st_mkint)(def, 2, True); break;
+      case -4:	type = VG_(st_mkint)(def, 4, True); break;
+      case -5:	type = VG_(st_mkint)(def, 1, False); break;
+      case -6:	type = VG_(st_mkint)(def, 1, True); break;
+      case -7:	type = VG_(st_mkint)(def, 2, False); break;
+      case -8:	type = VG_(st_mkint)(def, 4, False); break;
+      case -9:	type = VG_(st_mkint)(def, 4, False); break;
+      case -10:	type = VG_(st_mkint)(def, 4, False); break;
+      case -11:	type = VG_(st_mkvoid)(def); break;
+      case -12:	type = VG_(st_mkfloat)(def, 4); break;
+      case -13:	type = VG_(st_mkfloat)(def, 8); break;
+      case -15:	type = VG_(st_mkint)(def, 4, True); break;
+      case -16:	type = VG_(st_mkbool)(def, 4); break;
+      case -17:	type = VG_(st_mkfloat)(def, 4); break;
+      case -18:	type = VG_(st_mkfloat)(def, 8); break;
+      case -20:	type = VG_(st_mkint)(def, 1, False); break;
+      case -21:	type = VG_(st_mkint)(def, 1, False); break;
+      case -22:	type = VG_(st_mkint)(def, 2, False); break;
+      case -23:	type = VG_(st_mkint)(def, 4, False); break;
+      case -24:	type = VG_(st_mkint)(def, 4, False); break;
+      case -27:	type = VG_(st_mkint)(def, 1, True); break;
+      case -28:	type = VG_(st_mkint)(def, 2, True); break;
+      case -29:	type = VG_(st_mkint)(def, 4, True); break;
+      case -30:	type = VG_(st_mkint)(def, 2, False); break;
+      case -31:	type = VG_(st_mkint)(def, 8, True); break;
+      case -32:	type = VG_(st_mkint)(def, 8, False); break;
+      case -33:	type = VG_(st_mkint)(def, 8, False); break;
+      case -34:	type = VG_(st_mkint)(def, 8, True); break;
 
       default:
 	 VG_(printf)(" @@ unrecognized negative type %d\n", n);
 	 type = NULL;
 	 break;
       }
-      EXPECT(';', "negative builtin type");
+      /* Different versions of gcc seem to disagree about whether a
+         negative type is followed by a semicolon or not, and the stabs
+         spec (susch as it is) is not clear either so we will skip a
+         semicolon if there is one. */
+      if (*p == ';')
+         p++;
       break;
    }
 
@@ -601,10 +679,8 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
       EXPECT(';', "FP-TYPE");
       bytes = atoi(&p, 0);
       EXPECT(';', "FP-TYPE bytes");
-      atoi(&p, 0);
-      EXPECT(';', "FP-TYPE extra");
       
-      type = VG_(st_mkfloat)(def, bytes * 8);
+      type = VG_(st_mkfloat)(def, bytes);
       break;
    }
 
@@ -644,32 +720,26 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
 
    case '&':			/* reference */
    case '*': {			/* pointer */
-      /* '*' TYPE */
+      /* ('*' | '&') TYPE */
       type = stabtype_parser(si, NULL, &p);
       type = VG_(st_mkpointer)(def, type);
       break;
    }
 
+   case 'k':                    /* const */
+   case 'B':                    /* volatile */
+   case 'd': {                  /* file (pascal only) */
+      /* ('k' | 'B' | 'd' ) TYPE */
+      type = stabtype_parser(si, NULL, &p);
+      break;
+   }
+
    case 'x': {			/* reference to undefined type */
       /* 'x' ('s' | 'u' | 'e') NAME ':' */
-      Int brac = 0;		/* < > brackets in type */
       Char kind = *p++;		/* get kind */
       Char *name = p;
 
-      /* name is delimited by : except for :: within <> */
-      for(;;) {
-	 if (*p == '<')
-	    brac++;
-	 if (*p == '>')
-	    brac--;
-	 if (*p == ':') {
-	    if (brac && p[1] == ':')
-	       p += 1;
-	    else
-	       break;
-	 }
-	 p++;
-      }
+      p = templ_name(name);
       EXPECT(':', "struct/union/enum ref");
 
       name = VG_(addStr)(si, name, p-1-name);
@@ -692,9 +762,19 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
       break;
    }
 
+   case 'S': {			/* set/bitstring */
+      /* 'S' TYPE */
+      SymType *typeinfo;
+
+      typeinfo = stabtype_parser(si, NULL, &p);
+
+      type = VG_(st_mkarray)(def, typeinfo, VG_(st_mkint)(NULL, 1, True));
+      break;
+   }
+
    case 'P':			/* packed array */
    case 'a': {			/* array */
-      /* a IDX-TYPE TYPE */
+      /* ( 'a' | 'P' ) IDX-TYPE TYPE */
       SymType *idxtype;
       SymType *artype;
 
@@ -725,7 +805,7 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
    case 's': {			/* struct */
       /* Gad.  Here we go:
 
-	 's' SIZE
+	 ( 's' | 'u' ) SIZE
 		( '!' NBASE ',' ( VIRT PUB OFF ',' BASE-TYPE ){NBASE} )?
 
 		( NAME ( ':' ( '/' [0-9] )? TYPE ',' OFFSET ( ',' SIZE )?
@@ -769,8 +849,8 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
 	 UInt off, sz;
 	 SymType *fieldty;
 
-	 end = SKIPPAST(p, ':', "struct/union field NAME");
-	 end--;			/* want to point to (first) ':' */
+         end = SKIPPAST(p, ':', "method name") - 1;
+
 	 if (end[1] == ':') {
 	    /* c++ method names end in :: */
 	    method = True;
@@ -840,7 +920,10 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
 	    } else {
 	       EXPECT(',', "struct TYPE");
 
-	       off = atou(&p, 0);
+               /* logic dictates that the offset would always be
+	          positive and that atou would work here but GNAT has
+	          has other ideas - see bug 90128 for more details */
+	       off = atoi(&p, 0);
 
 	       if (*p == ',') {
 		  EXPECT(',', "struct OFFSET");
@@ -907,6 +990,15 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
       EXPECT(';', "method definition");
       break;
 
+   case '@':			/* pointer to member */
+      /* '@' CLASS-TYPE ',' MEMBER-TYPE */
+      type = VG_(st_mkint)(def, sizeof(int), False); /* make it an int for our use */
+      
+      stabtype_parser(si, NULL, &p); /* CLASS-TYPE */
+      EXPECT(',', "member-pointer CLASS-TYPE");
+      stabtype_parser(si, NULL, &p); /* MEMBER-TYPE */
+      break;
+
    default:
       VG_(printf)(" @@ don't know what type '%c' is\n", t);
       type = NULL;
@@ -926,7 +1018,7 @@ static SymType *stabtype_parser(SegInfo *si, SymType *def, Char **pp)
 /* parse a symbol reference: NAME ':' DESC TYPE */
 static Bool initSym(SegInfo *si, Sym *sym, stab_types kind, Char **namep, Int val)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    Char *name = *namep;
    Char *ty;
    Int len;
@@ -938,7 +1030,18 @@ static Bool initSym(SegInfo *si, Sym *sym, stab_types kind, Char **namep, Int va
       VG_(printf)("initSym(si=%p, tab=%p, sym=%p, kind=%d, name=%p \"%s\", val=%d)\n",
 		  si, si->stab_typetab, sym, kind, name, name, val);
 
-   ty = nested_name(name);
+   /* Find first ':' */
+   ty = VG_(strchr)(name, ':');
+
+   /* Skip '::' */
+   while (ty && ty[1] == ':')
+      ty = VG_(strchr)(ty + 2, ':');
+
+   if (ty == NULL) {
+      /* there was no ':' */
+      *namep += VG_(strlen)(name);
+      return True;		/* skip */
+   }
 
    len = ty - name;
 
@@ -984,7 +1087,6 @@ static Bool initSym(SegInfo *si, Sym *sym, stab_types kind, Char **namep, Int va
      out:
       sym->type = stabtype_parser(si, NULL, &ty);
       base = VG_(st_basetype)(sym->type, False);
-
       if (isStruct && (VG_(st_isstruct)(base) || VG_(st_isunion)(base))) {
 	 Char *sname = VG_(addStr)(si, name, len);
 	 structDef(si->stab_typetab, base, VG_(st_isstruct)(base), sname);
@@ -1005,12 +1107,12 @@ static Bool initSym(SegInfo *si, Sym *sym, stab_types kind, Char **namep, Int va
    case N_STSYM:
    case N_LCSYM:
       sym->kind = SyStatic;
-      sym->addr = si->offset + (Addr)val;
+      sym->u.addr = si->offset + (Addr)val;
       break;
 
    case N_PSYM:
       sym->kind = SyEBPrel;	/* +ve offset off EBP (erk, or ESP if no frame pointer) */
-      sym->offset = val;
+      sym->u.offset = val;
       break;
 
    case N_LSYM:
@@ -1018,17 +1120,17 @@ static Bool initSym(SegInfo *si, Sym *sym, stab_types kind, Char **namep, Int va
 	 sym->kind = SyEBPrel;	/* -ve off EBP when there's a frame pointer */
       else
 	 sym->kind = SyESPrel;	/* +ve off ESP when there's no frame pointer */
-      sym->offset = val;
+      sym->u.offset = val;
       break;
 
    case N_RSYM:
       sym->kind = SyReg;
-      sym->regno = val;
+      sym->u.regno = val;
       break;
 
    case N_GSYM:
       sym->kind = SyGlobal;
-      sym->addr = 0;		/* XXX should really look up global address */
+      sym->u.addr = 0;		/* XXX should really look up global address */
       break;
 
    default:
@@ -1057,7 +1159,7 @@ struct symlist {
    definitions helps a lot. */
 static Scope *addSymsToScope(Scope *sc, struct symlist *list, Int nsyms, Scope *outer)
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    Int j;
    struct symlist *n;
    Int base;
@@ -1103,7 +1205,7 @@ void VG_(read_debuginfo_stabs) ( SegInfo* si,
 				 UChar* stabC,   Int stab_sz, 
 				 UChar* stabstr, Int stabstr_sz )
 {
-   static const Bool debug = False;
+   static const Bool debug = False || stabs_debug;
    Int    i;
    Int    n_stab_entries;
    struct nlist* stab = (struct nlist*)stabC;
@@ -1181,7 +1283,7 @@ void VG_(read_debuginfo_stabs) ( SegInfo* si,
 
       /* handle continued string stabs */
       {
-	 static const Bool contdebug = False;
+	 static const Bool contdebug = False || stabs_debug;
 	 Int buflen = 0;
 	 Int idx = 0;
 	 Char *buf = NULL;
@@ -1595,5 +1697,5 @@ void VG_(read_debuginfo_stabs) ( SegInfo* si,
 }
 
 /*--------------------------------------------------------------------*/
-/*--- end                                        vg_symtab_stabs.c ---*/
+/*--- end                                                          ---*/
 /*--------------------------------------------------------------------*/
