@@ -6,7 +6,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward
+   Copyright (C) 2000-2005 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -27,7 +27,7 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "vg_include.h"
+#include "core.h"
 #include "vg_symtypes.h"
 
 typedef enum {
@@ -399,6 +399,7 @@ SymType *VG_(st_mktypedef)(SymType *st, Char *name, SymType *type)
 {
    st = alloc(st);
 
+   vg_assert(st != type);
    vg_assert(st->kind == TyUnresolved || st->kind == TyUnknown ||
 	     st->kind == TyStruct || st->kind == TyUnion ||
 	     st->kind == TyTypedef);
@@ -500,12 +501,12 @@ static void clear_visited(void)
    }
 }
 
-static void bprintf(void (*send)(Char), const Char *fmt, ...)
+static void bprintf(void (*send)(Char, void*), void *send_arg, const Char *fmt, ...)
 {
    va_list vargs;
 
    va_start(vargs, fmt);
-   VG_(vprintf)(send, fmt, vargs);
+   VG_(vprintf)(send, fmt, vargs, send_arg);
    va_end(vargs);
 }
 
@@ -521,9 +522,9 @@ static ShadowChunk *findchunk(Addr a)
 }
 #endif
 
-static vki_ksigaction sigbus_saved;
-static vki_ksigaction sigsegv_saved;
-static vki_ksigset_t  blockmask_saved;
+static struct vki_sigaction sigbus_saved;
+static struct vki_sigaction sigsegv_saved;
+static vki_sigset_t  blockmask_saved;
 static jmp_buf valid_addr_jmpbuf;
 
 static void valid_addr_handler(int sig)
@@ -537,39 +538,39 @@ static void valid_addr_handler(int sig)
 static void setup_signals(void)
 {
    Int res;
-   vki_ksigaction sigbus_new;
-   vki_ksigaction sigsegv_new;
-   vki_ksigset_t  unblockmask_new;
+   struct vki_sigaction sigbus_new;
+   struct vki_sigaction sigsegv_new;
+   vki_sigset_t         unblockmask_new;
 
    /* Temporarily install a new sigsegv and sigbus handler, and make
       sure SIGBUS, SIGSEGV and SIGTERM are unblocked.  (Perhaps the
       first two can never be blocked anyway?)  */
 
    sigbus_new.ksa_handler = valid_addr_handler;
-   sigbus_new.ksa_flags = VKI_SA_ONSTACK | VKI_SA_RESTART;
-   sigbus_new.ksa_restorer = NULL;
-   res = VG_(ksigemptyset)( &sigbus_new.ksa_mask );
+   sigbus_new.sa_flags = VKI_SA_ONSTACK | VKI_SA_RESTART;
+   sigbus_new.sa_restorer = NULL;
+   res = VG_(sigemptyset)( &sigbus_new.sa_mask );
    vg_assert(res == 0);
 
    sigsegv_new.ksa_handler = valid_addr_handler;
-   sigsegv_new.ksa_flags = VKI_SA_ONSTACK | VKI_SA_RESTART;
-   sigsegv_new.ksa_restorer = NULL;
-   res = VG_(ksigemptyset)( &sigsegv_new.ksa_mask );
+   sigsegv_new.sa_flags = VKI_SA_ONSTACK | VKI_SA_RESTART;
+   sigsegv_new.sa_restorer = NULL;
+   res = VG_(sigemptyset)( &sigsegv_new.sa_mask );
    vg_assert(res == 0+0);
 
-   res =  VG_(ksigemptyset)( &unblockmask_new );
-   res |= VG_(ksigaddset)( &unblockmask_new, VKI_SIGBUS );
-   res |= VG_(ksigaddset)( &unblockmask_new, VKI_SIGSEGV );
-   res |= VG_(ksigaddset)( &unblockmask_new, VKI_SIGTERM );
+   res =  VG_(sigemptyset)( &unblockmask_new );
+   res |= VG_(sigaddset)( &unblockmask_new, VKI_SIGBUS );
+   res |= VG_(sigaddset)( &unblockmask_new, VKI_SIGSEGV );
+   res |= VG_(sigaddset)( &unblockmask_new, VKI_SIGTERM );
    vg_assert(res == 0+0+0);
 
-   res = VG_(ksigaction)( VKI_SIGBUS, &sigbus_new, &sigbus_saved );
+   res = VG_(sigaction)( VKI_SIGBUS, &sigbus_new, &sigbus_saved );
    vg_assert(res == 0+0+0+0);
 
-   res = VG_(ksigaction)( VKI_SIGSEGV, &sigsegv_new, &sigsegv_saved );
+   res = VG_(sigaction)( VKI_SIGSEGV, &sigsegv_new, &sigsegv_saved );
    vg_assert(res == 0+0+0+0+0);
 
-   res = VG_(ksigprocmask)( VKI_SIG_UNBLOCK, &unblockmask_new, &blockmask_saved );
+   res = VG_(sigprocmask)( VKI_SIG_UNBLOCK, &unblockmask_new, &blockmask_saved );
    vg_assert(res == 0+0+0+0+0+0);
 }
 
@@ -578,13 +579,13 @@ static void restore_signals(void)
    Int res;
 
    /* Restore signal state to whatever it was before. */
-   res = VG_(ksigaction)( VKI_SIGBUS, &sigbus_saved, NULL );
+   res = VG_(sigaction)( VKI_SIGBUS, &sigbus_saved, NULL );
    vg_assert(res == 0 +0);
 
-   res = VG_(ksigaction)( VKI_SIGSEGV, &sigsegv_saved, NULL );
+   res = VG_(sigaction)( VKI_SIGSEGV, &sigsegv_saved, NULL );
    vg_assert(res == 0 +0 +0);
 
-   res = VG_(ksigprocmask)( VKI_SIG_SETMASK, &blockmask_saved, NULL );
+   res = VG_(sigprocmask)( VKI_SIG_SETMASK, &blockmask_saved, NULL );
    vg_assert(res == 0 +0 +0 +0);
 }
 
@@ -597,8 +598,7 @@ static Bool is_valid_addr(Addr a)
    static const Bool debug = False;
    volatile Bool ret = False;
 
-   if ((a > VKI_BYTES_PER_PAGE) &&
-       !test_visited(a, &faulted)) {
+   if ((a > VKI_PAGE_SIZE) && !test_visited(a, &faulted)) {
       if (!LAZYSIG)
 	 setup_signals();
    
@@ -667,7 +667,7 @@ static UInt describe_addr_bufidx;
 static UInt describe_addr_bufsz;
 
 /* Add a character to the result buffer */
-static void describe_addr_addbuf(Char c) {
+static void describe_addr_addbuf(Char c,void *p) {
    if ((describe_addr_bufidx+1) >= describe_addr_bufsz) {
       Char *n;
     
@@ -872,7 +872,7 @@ Char *VG_(describe_addr)(ThreadId tid, Addr addr)
 	       if (debug)
 		  VG_(printf)("    non-followable array (sz=%d): checking addr %p in range %p-%p\n",
 			      sz, addr, var->valuep, (var->valuep + sz));
-	       if (addr >= var->valuep && addr <= (var->valuep + sz))
+	       if (ty->size > 0 && addr >= var->valuep && addr <= (var->valuep + sz))
 		  min = max = (addr - var->valuep) / ty->size;
 	       else
 		  break;
@@ -1018,7 +1018,7 @@ Char *VG_(describe_addr)(ThreadId tid, Addr addr)
 	       found->container->name = NULL;
 	       found->container = found->container->container;
 	    } else {
-	       bprintf(describe_addr_addbuf, "&(");
+               bprintf(describe_addr_addbuf, 0, "&(");
 	       ptr = False;
 	    }
 
@@ -1030,13 +1030,13 @@ Char *VG_(describe_addr)(ThreadId tid, Addr addr)
 
 	 *ep++ = '\0';
 
-	 bprintf(describe_addr_addbuf, sp);
+	 bprintf(describe_addr_addbuf, 0, sp);
 
 	 if (addr != found->valuep)
-	    bprintf(describe_addr_addbuf, "+%d", addr - found->valuep);
+	    bprintf(describe_addr_addbuf, 0, "+%d", addr - found->valuep);
 
 	 if (VG_(get_filename_linenum)(eip, file, sizeof(file), &line))
-	    bprintf(describe_addr_addbuf, " at %s:%d", file, line, addr);
+	    bprintf(describe_addr_addbuf, 0, " at %s:%d", file, line, addr);
       }
    }
 
