@@ -8,7 +8,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -29,7 +29,7 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "vg_include.h"
+#include "core.h"
 
 /*------------------------------------------------------------*/
 /*--- Renamings of frequently-used global functions.       ---*/
@@ -71,11 +71,17 @@ void VG_(print_reg_alloc_stats)(void)
 static UCodeBlock* alloc_UCodeBlock ( Addr orig_eip )
 {
    UCodeBlock* cb = VG_(arena_malloc)(VG_AR_CORE, sizeof(UCodeBlock));
+   Int i;
+
    cb->orig_eip = orig_eip;
    cb->used     = 0;
    cb->size     = 0;
    cb->instrs   = NULL;
    cb->nextTemp = 0;
+   cb->relocs   = NULL;
+   cb->nrelocs  = 0;
+   for(i = 0; i < VG_MAX_JUMPS; i++)
+      cb->jumps[i] = -1;
    return cb;
 }
 
@@ -90,6 +96,7 @@ UCodeBlock* VG_(setup_UCodeBlock) ( UCodeBlock* cb_in )
 void VG_(free_UCodeBlock) ( UCodeBlock* cb )
 {
    if (cb->instrs) VG_(arena_free)(VG_AR_CORE, cb->instrs);
+   if (cb->relocs) VG_(arena_free)(VG_AR_JITTER, cb->relocs);
    VG_(arena_free)(VG_AR_CORE, cb);
 }
 
@@ -284,7 +291,7 @@ Bool VG_(any_flag_use) ( UInstr* u )
 /* Convert a rank in the range 0 .. VG_MAX_REALREGS-1 into an Intel
    register number.  This effectively defines the order in which real
    registers are allocated.  %ebp is excluded since it is permanently
-   reserved for pointing at VG_(baseBlock).
+   reserved for pointing at the thread's ThreadState.
 
    Important!  This function must correspond with the value of
    VG_MAX_REALREGS (actually, VG_MAX_REALREGS can be reduced without
@@ -547,7 +554,7 @@ static Bool is_sane_UInstr ( UInstr* u )
 
    /* Fields checked: lit32   size  flags_r/w tag1   tag2   tag3    (rest) */
    case PUTSEG: return LIT0 && SZ2  && CC0 &&  TR1 && Se2 &&  N3 && XOTHER;
-   case GETSEG: return LIT0 && SZ2  && CC0 &&  Se1 && TR2 &&  N3 && XOTHER;
+   case GETSEG: return LIT0 && SZ42 && CC0 &&  Se1 && TR2 &&  N3 && XOTHER;
    case USESEG: return LIT0 && SZ0  && CC0 &&  TR1 && TR2 &&  N3 && XOTHER;
    case NOP:    return LIT0 && SZ0  && CC0 &&   N1 &&  N2 &&  N3 && XOTHER;
    case LOCK:   return LIT0 && SZ0  && CC0 &&   N1 &&  N2 &&  N3 && XOTHER;
@@ -561,7 +568,7 @@ static Bool is_sane_UInstr ( UInstr* u )
    case CMOV:   return LIT0 && SZ4  && CCg &&  TR1 && TR2 &&  N3 && XCONDi;
    case WIDEN:  return LIT0 && SZ42 && CC0 &&  TR1 &&  N2 &&  N3 && XWIDEN;
    case JMP:    return LITm && SZ0  && CCj && TRL1 &&  N2 &&  N3 && XJMP;
-   case CALLM:  return LIT0 && SZ0 /*any*/ &&  Ls1 &&  N2 &&  N3 && XOTHER;
+   case CALLM:  return LIT1 && SZ0 /*any*/ &&   L1 &&  N2 &&  N3 && XOTHER;
    case CALLM_S: 
    case CALLM_E:return LIT0 && SZ0  && CC0 &&   N1 &&  N2 &&  N3 && XOTHER;
    case PUSH: 
@@ -1589,8 +1596,8 @@ Bool uInstrMentionsTempReg ( UInstr* u, Int tempreg )
 static void vg_improve ( UCodeBlock* cb )
 {
    Int     i, j, k, m, n, ar, tr, told, actual_areg;
-   Int     areg_map[8];
-   Bool    annul_put[8];
+   Int     areg_map[N_ARCH_REGS];
+   Bool    annul_put[N_ARCH_REGS];
    Int     tempUse[VG_MAX_REGS_USED];
    Bool    isWrites[VG_MAX_REGS_USED];
    UInstr* u;
@@ -1636,14 +1643,14 @@ static void vg_improve ( UCodeBlock* cb )
 #  define BIND_ARCH_TO_TEMP(archreg,tempreg)\
    { Int q;                                           \
      /* Invalidate any old binding(s) to tempreg. */  \
-     for (q = 0; q < 8; q++)                          \
+     for (q = 0; q < N_ARCH_REGS; q++)                \
         if (areg_map[q] == tempreg) areg_map[q] = -1; \
      /* Add the new binding. */                       \
      areg_map[archreg] = (tempreg);                   \
    }
 
    /* Set up the A-reg map. */
-   for (i = 0; i < 8; i++) areg_map[i] = -1;
+   for (i = 0; i < N_ARCH_REGS; i++) areg_map[i] = -1;
 
    /* Scan insns. */
    for (i = 0; i < cb->used; i++) {
@@ -1735,7 +1742,7 @@ static void vg_improve ( UCodeBlock* cb )
             wr  = isWrites[j];
             if (!wr) continue;
             tr = tempUse[j];
-            for (m = 0; m < 8; m++)
+            for (m = 0; m < N_ARCH_REGS; m++)
                if (areg_map[m] == tr) areg_map[m] = -1;
          }
       }
@@ -1749,7 +1756,7 @@ static void vg_improve ( UCodeBlock* cb )
       in-memory value of %ESP to be up to date.  Although this isn't
       actually required by other analyses (cache simulation), it's
       simplest to be consistent for all end-uses. */
-   for (j = 0; j < 8; j++)
+   for (j = 0; j < N_ARCH_REGS; j++)
       annul_put[j] = False;
 
    for (i = cb->used-1; i >= 0; i--) {
@@ -1775,7 +1782,7 @@ static void vg_improve ( UCodeBlock* cb )
       } 
       else if (u->opcode == JMP || u->opcode == JIFZ
                || u->opcode == CALLM) {
-         for (j = 0; j < 8; j++)
+         for (j = 0; j < N_ARCH_REGS; j++)
             annul_put[j] = False;
       }
       else {
@@ -1930,7 +1937,7 @@ UCodeBlock* vg_ESP_update_pass(UCodeBlock* cb_in)
                if (VG_(defined_new_mem_stack)() ||		\
                    VG_(defined_die_mem_stack)()) {		\
                   uInstr1(cb, CCALL, 0, TempReg, u->val1);	\
-                  uCCall(cb, (Addr) VG_(unknown_esp_update),	\
+                  uCCall(cb, (Addr) VG_(unknown_SP_update),	\
                          1, 1, False);				\
                } 
 
@@ -2428,12 +2435,11 @@ static void vg_realreg_liveness_analysis ( UCodeBlock* cb )
 
    'tid' is the identity of the thread needing this block.
 */
-void VG_(translate) ( ThreadId tid, Addr orig_addr,
+Bool VG_(translate) ( ThreadId tid, Addr orig_addr,
                       Bool debugging_translation )
 {
    Addr        trans_addr, redir, orig_addr0 = orig_addr;
-   UShort      jumps[VG_MAX_JUMPS];
-   Int         i, orig_size, trans_size;
+   Int         orig_size, trans_size;
    UCodeBlock* cb;
    Bool        notrace_until_done;
    UInt        notrace_until_limit = 0;
@@ -2444,9 +2450,6 @@ void VG_(translate) ( ThreadId tid, Addr orig_addr,
    beforeRA        = True;
    beforeLiveness  = True;
 
-   for (i = 0; i < VG_MAX_JUMPS; i++)
-      jumps[i] = (UShort)-1;
-
    /* Look in the code redirect table to see if we should
       translate an alternative address for orig_addr. */
    redir = VG_(code_redirect)(orig_addr);
@@ -2467,24 +2470,21 @@ void VG_(translate) ( ThreadId tid, Addr orig_addr,
       before --trace-codegen= style printing takes effect. */
    notrace_until_done = VG_(get_bbs_translated)() >= notrace_until_limit;
 
-   seg = VG_(find_segment)(orig_addr);
+   seg = VG_(find_segment_containing)(orig_addr);
 
    if (!debugging_translation)
       VG_TRACK( pre_mem_read, Vg_CoreTranslate, tid, "", orig_addr, 1 );
 
    if (seg == NULL ||
-       !VG_(seg_contains)(seg, orig_addr, 1) || 
        (seg->prot & (VKI_PROT_READ|VKI_PROT_EXEC)) == 0) {
       /* Code address is bad - deliver a signal instead */
-      vg_assert(!VG_(is_addressable)(orig_addr, 1));
-
-      if (seg != NULL && VG_(seg_contains)(seg, orig_addr, 1)) {
+      if (seg != NULL) {
 	 vg_assert((seg->prot & VKI_PROT_EXEC) == 0);
 	 VG_(synth_fault_perms)(tid, orig_addr);
       } else
 	 VG_(synth_fault_mapping)(tid, orig_addr);
 
-      return;
+      return False;
    } else
       seg->flags |= SF_CODE;	/* contains cached code */
 
@@ -2536,7 +2536,7 @@ void VG_(translate) ( ThreadId tid, Addr orig_addr,
 
    /* Add %ESP-update hooks if the tool requires them */
    /* Nb: We don't print out this phase, because it doesn't do much */
-   if (VG_(need_to_handle_esp_assignment)()) {
+   if (VG_(need_to_handle_SP_assignment)()) {
       VGP_PUSHCC(VgpESPUpdate);
       cb = vg_ESP_update_pass ( cb );
       VGP_POPCC(VgpESPUpdate);
@@ -2559,9 +2559,8 @@ void VG_(translate) ( ThreadId tid, Addr orig_addr,
    /* Emit final code */
    VG_(print_codegen) = DECIDE_IF_PRINTING_CODEGEN_FOR_PHASE(5);
    VGP_PUSHCC(VgpFromUcode);
-   trans_addr = (Addr)VG_(emit_code)(cb, &trans_size, jumps );
+   trans_addr = (Addr)VG_(emit_code)(cb, &trans_size );
    VGP_POPCC(VgpFromUcode);
-   VG_(free_UCodeBlock)(cb);
 
 #undef DECIDE_IF_PRINTING_CODEGEN_FOR_PHASE
 
@@ -2575,14 +2574,16 @@ void VG_(translate) ( ThreadId tid, Addr orig_addr,
    if (!debugging_translation) {
       // Note that we use orig_addr0, not orig_addr, which might have been
       // changed by the redirection
-      VG_(add_to_trans_tab)( orig_addr0, orig_size, trans_addr, trans_size,
-                             jumps );
+      VG_(add_to_trans_tab)( orig_addr0, orig_size, trans_addr, trans_size, cb );
    }
 
    /* Free the intermediary -- was allocated by VG_(emit_code). */
    VG_(arena_free)( VG_AR_JITTER, (void*)trans_addr );
+   VG_(free_UCodeBlock)(cb);
 
    VGP_POPCC(VgpTranslate);
+
+   return True;
 }
 
 
