@@ -8,7 +8,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -29,8 +29,8 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "vg_include.h"
-#include <stddef.h>
+#include "core.h"
+#include "memcheck/memcheck.h"
 
 /* #define DEBUG_TRANSTAB */
 
@@ -59,22 +59,10 @@ static /* const */ Int vg_tc_sector_szB = 0;
 
 /*------------------ TYPES ------------------*/
 
-#define CODE_ALIGNMENT	4	/* alignment of TCEntries */
+#define CODE_ALIGNMENT	sizeof(void*)     // alignment of TCEntries
 #define CODE_ALIGN(a)	(((a)+CODE_ALIGNMENT-1) & ~(CODE_ALIGNMENT-1))
 #define IS_ALIGNED(a)	(((a) & (CODE_ALIGNMENT-1)) == 0)
 
-/* An entry in TC.  Payload always is always padded out to a 4-aligned
-   quantity so that these structs are always word-aligned. */
-typedef
-   struct { 
-      /* +0 */ Addr   orig_addr;
-      /* +4 */ UShort orig_size;
-      /* +6 */ UShort trans_size;
-      /* +8 */ UShort jump_sites[VG_MAX_JUMPS];
-      /* +VG_CODE_OFFSET */ UChar  payload[0];
-   }
-   TCEntry;
-
 /* An entry in TT. */
 typedef
    struct {
@@ -238,7 +226,7 @@ void add_tt_entry ( TCEntry* tce )
    UInt i;
    /* VG_(printf)("add_TT_entry orig_addr %p\n", tce->orig_addr); */
    /* Hash to get initial probe point. */
-   i = ((UInt)(tce->orig_addr)) % VG_TT_SIZE;
+   i = tce->orig_addr % VG_TT_SIZE;
    while (True) {
       if (vg_tt[i].orig_addr == tce->orig_addr)
          VG_(core_panic)("add_TT_entry: duplicate");
@@ -266,7 +254,7 @@ TTEntry* search_tt ( Addr orig_addr )
 {
    Int i;
    /* Hash to get initial probe point. */
-   i = ((UInt)orig_addr) % VG_TT_SIZE;
+   i = orig_addr % VG_TT_SIZE;
    while (True) {
       if (vg_tt[i].orig_addr == orig_addr)
          return &vg_tt[i];
@@ -360,11 +348,11 @@ void discard_oldest_sector ( void )
       VG_(sprintf)(msg, "before discard of sector %d (%d bytes)", 
                         s, vg_tc_used[s]);
 
+      VALGRIND_DISCARD_TRANSLATIONS(vg_tc[s], vg_tc_used[s]);
       for(i = 0; i < VG_TC_N_SECTORS; i++) {
 	 if (i != s && vg_tc[i] != NULL)
 	    unchain_sector(i, (Addr)vg_tc[s], vg_tc_used[s]);
       }
-
       pp_tt_tc_status ( msg );
       overall_out_count += vg_tc_stats_count[s];
       overall_out_osize += vg_tc_stats_osize[s];
@@ -400,27 +388,29 @@ Int maybe_commission_sector ( void )
    }
    for (s = 0; s < VG_TC_N_SECTORS; s++) {
       if (vg_tc[s] == NULL) {
-#if 1
-         vg_tc[s] = VG_(get_memory_from_mmap) 
-                       ( vg_tc_sector_szB, "trans-cache(sector)" );
-#else
-         // Alternative: put translations in an mmap'd file.  The main
-         // reason is to help OProfile -- OProfile can assign time spent in
-         // translations to a particular file.  The file format doesn't
-         // really matter, which is good because it's not really readable,
-         // being generated code but not a proper ELF file.
-	 Char buf[20];
-	 static Int count = 0;
-	 Int fd;
+	 if (1)
+	    vg_tc[s] = VG_(get_memory_from_mmap) 
+	       ( vg_tc_sector_szB, "trans-cache(sector)" );
+	 else {
+	    // Alternative: put translations in an mmap'd file.  The main
+	    // reason is to help OProfile -- OProfile can assign time spent in
+	    // translations to a particular file.  The file format doesn't
+	    // really matter, which is good because it's not really readable,
+	    // being generated code but not a proper ELF file.
+	    Char buf[20];
+	    static Int count = 0;
+	    Int fd;
 	 
-	 VG_(sprintf)(buf, ".transtab.%d", count++);
+	    VG_(sprintf)(buf, ".transtab.%d", count++);
 
-	 fd = VG_(open)(buf, VKI_O_RDWR|VKI_O_CREAT|VKI_O_TRUNC, 0700);
-	 //VG_(unlink)(buf);
-	 VG_(do_syscall)(__NR_ftruncate, fd, PGROUNDUP(vg_tc_sector_szB));
-	 vg_tc[s] = VG_(mmap)(0, PGROUNDUP(vg_tc_sector_szB), VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC, VKI_MAP_SHARED, 0, fd, 0);
-	 VG_(close)(fd);
-#endif
+	    fd = VG_(open)(buf, VKI_O_RDWR|VKI_O_CREAT|VKI_O_TRUNC, 0700);
+	    //VG_(unlink)(buf);
+	    VG_(do_syscall)(__NR_ftruncate, fd, PGROUNDUP(vg_tc_sector_szB));
+	    vg_tc[s] = VG_(mmap)(0, PGROUNDUP(vg_tc_sector_szB), VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC, VKI_MAP_SHARED, 0, fd, 0);
+	    VG_(close)(fd);
+	 }
+
+	 VALGRIND_MAKE_WRITABLE(vg_tc[s], vg_tc_sector_szB);
          vg_tc_used[s] = 0;
          VG_(sprintf)(msg, "after  allocation of sector %d (size %d)", 
                            s, vg_tc_sector_szB );
@@ -555,10 +545,12 @@ Int VG_(get_bbs_translated) ( void )
 */
 void VG_(add_to_trans_tab) ( Addr orig_addr,  Int orig_size,
                              Addr trans_addr, Int trans_size,
-			     UShort jumps[VG_MAX_JUMPS])
+			     UCodeBlock *cb )
 {
    Int i, nBytes, trans_size_aligned;
    TCEntry* tce;
+   UShort prev;
+
    /*
    VG_(printf)("add_to_trans_tab(%d) %x %d %x %d\n",
                vg_tt_used, tte->orig_addr, tte->orig_size, 
@@ -583,13 +575,28 @@ void VG_(add_to_trans_tab) ( Addr orig_addr,  Int orig_size,
    tce->orig_size  = (UShort)orig_size;  /* what's the point of storing this? */
    tce->trans_size = (UShort)trans_size_aligned;
    for (i = 0; i < VG_MAX_JUMPS; i++) {
-      tce->jump_sites[i] = jumps[i];
+      tce->jump_sites[i] = cb->jumps[i];
    }
    for (i = 0; i < trans_size; i++) {
       tce->payload[i] = ((UChar*)trans_addr)[i];
    }
-   
-   unchain_tce(tce);
+
+   if (VG_(clo_trace_codegen) & 1)
+      VG_(message)(Vg_DebugMsg, "add_to_trans_tab: inserting code for %p into transtab at %p\n",
+		   orig_addr, tce->payload);
+
+   /* relocate absolute jumps */
+   prev = 0;
+   for(i = 0; i < cb->nrelocs; i++) {
+      UShort off = cb->relocs[i];
+
+      vg_assert(off < trans_size);
+      vg_assert(i == 0 || off > prev);
+
+      VG_(reloc_abs_jump)(&tce->payload[off]);
+
+      prev = off;
+   }
    add_tt_entry(tce);
 
    /* Update stats. */
@@ -610,6 +617,10 @@ void VG_(add_to_trans_tab) ( Addr orig_addr,  Int orig_size,
 Addr VG_(search_transtab) ( Addr original_addr )
 {
    TTEntry* tte;
+
+   if (vg_tt == NULL)
+      return 0;
+
    VGP_PUSHCC(VgpSlowFindT);
    tte = search_tt ( original_addr );
    if (tte == NULL) {
@@ -639,6 +650,10 @@ void VG_(invalidate_translations) ( Addr start, UInt range, Bool unchain_blocks
    UInt     out_count, out_osize, out_tsize;
    Int      i, j;
    TCEntry* tce;
+
+   if (vg_tt == NULL)
+      return;
+
 #  ifdef DEBUG_TRANSTAB
    VG_(sanity_check_tt_tc)();
 #  endif
@@ -662,7 +677,7 @@ void VG_(invalidate_translations) ( Addr start, UInt range, Bool unchain_blocks
       vg_tt[i].orig_addr = VG_TTE_DELETED;
       tce->orig_addr = VG_TTE_DELETED;
 
-      if (unchain_blocks) {
+      if (VG_(clo_chain_bb) && unchain_blocks) {
          /* make sure no other blocks chain to the one we just discarded */
          for(j = 0; j < VG_TC_N_SECTORS; j++) {
             if (vg_tc[j] != NULL)
@@ -708,6 +723,9 @@ void VG_(init_tt_tc) ( void )
       TCEntries. */
    vg_assert((VG_MAX_JUMPS % 2) == 0);
 
+   // Otherwise lots of things go wrong... 
+   vg_assert(VG_CODE_OFFSET == sizeof(TCEntry));
+   
    /* Figure out how big each sector should be.  */
    vg_tc_sector_szB 
       = (VG_TT_LIMIT /* max TT entries we expect */
