@@ -9,7 +9,7 @@
    detecting memory errors, and AddrCheck, a lightweight Valgrind tool 
    for detecting memory errors.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -36,7 +36,7 @@
 #ifndef __MAC_SHARED_H
 #define __MAC_SHARED_H
 
-#include "vg_skin.h"
+#include "tool.h"
 
 #define MAC_(str)    VGAPPEND(vgMAC_,str)
 
@@ -47,12 +47,13 @@
 /* The classification of a faulting address. */
 typedef 
    enum { 
-      Undescribed,  /* as-yet unclassified */
+      Undescribed,   // as-yet unclassified
       Stack, 
-      Unknown,      /* classification yielded nothing useful */
+      Unknown,       // classification yielded nothing useful
       Freed, Mallocd, 
-      UserG,        /* in a user-defined block; Addrcheck & Memcheck only */
-      Mempool,      /* in a mempool; Addrcheck & Memcheck only */
+      UserG,         // in a user-defined block
+      Mempool,       // in a mempool
+      Register,      // in a register;  for Param errors only
    }
    AddrKind;
 
@@ -60,10 +61,11 @@ typedef
 typedef
    struct {                   // Used by:
       AddrKind akind;         //   ALL
-      Int blksize;            //   Freed, Mallocd
-      Int rwoffset;           //   Freed, Mallocd
+      SizeT blksize;          //   Freed, Mallocd
+      OffT rwoffset;          //   Freed, Mallocd
       ExeContext* lastchange; //   Freed, Mallocd
       ThreadId stack_tid;     //   Stack
+      const Char *desc;	      //   UserG
       Bool maybe_gcc;         // True if just below %esp -- could be a gcc bug.
    }
    AddrInfo;
@@ -110,7 +112,7 @@ typedef
       AxsKind axskind;     //   AddrErr
       Int size;            //   AddrErr, ValueErr
       AddrInfo addrinfo;   //   {Addr,Free,FreeMismatch,Param,User}Err
-      Bool isWrite;        //   ParamErr, UserErr, CoreMemErr
+      Bool isUnaddr;       //   {CoreMem,Param,User}Err
    }
    MAC_Error;
 
@@ -191,13 +193,28 @@ extern UInt MAC_(event_ctr)[N_PROF_EVENTS];
 /*--- V and A bits                                         ---*/
 /*------------------------------------------------------------*/
 
-#define IS_DISTINGUISHED_SM(smap) \
-   ((smap) == &distinguished_secondary_map)
+/* expand 1 bit -> 8 */
+#define BIT_EXPAND(b)	((~(((UChar)(b) & 1) - 1)) & 0xFF)
+
+#define SECONDARY_SHIFT	16
+#define SECONDARY_SIZE	(1 << SECONDARY_SHIFT)
+#define SECONDARY_MASK	(SECONDARY_SIZE - 1)
+
+#define PRIMARY_SIZE	(1 << (32 - SECONDARY_SHIFT))
+
+#define SM_OFF(addr)	((addr) & SECONDARY_MASK)
+#define PM_IDX(addr)	((addr) >> SECONDARY_SHIFT)
+
+#define IS_DISTINGUISHED_SM(smap)		   \
+   ((smap) >= &distinguished_secondary_maps[0] &&  \
+    (smap) < &distinguished_secondary_maps[N_SECONDARY_MAPS])
+
+#define IS_DISTINGUISHED(addr)	(IS_DISTINGUISHED_SM(primary_map[PM_IDX(addr)]))
 
 #define ENSURE_MAPPABLE(addr,caller)                              \
    do {                                                           \
-      if (IS_DISTINGUISHED_SM(primary_map[(addr) >> 16])) {       \
-         primary_map[(addr) >> 16] = alloc_secondary_map(caller); \
+      if (IS_DISTINGUISHED(addr)) {				  \
+	 primary_map[PM_IDX(addr)] = alloc_secondary_map(caller, primary_map[PM_IDX(addr)]); \
          /* VG_(printf)("new 2map because of %p\n", addr); */     \
       }                                                           \
    } while(0)
@@ -249,7 +266,15 @@ extern Bool MAC_(clo_partial_loads_ok);
 extern Int MAC_(clo_freelist_vol);
 
 /* Do leak check at exit?  default: NO */
-extern Bool MAC_(clo_leak_check);
+typedef
+   enum {
+      LC_Off,
+      LC_Summary,
+      LC_Full,
+   }
+   LeakCheckMode;
+
+extern LeakCheckMode MAC_(clo_leak_check);
 
 /* How closely should we compare ExeContexts in leak records? default: 2 */
 extern VgRes MAC_(clo_leak_resolution);
@@ -277,19 +302,20 @@ extern VgHashTable MAC_(malloc_list);
 extern VgHashTable MAC_(mempool_list);
 
 /* Function pointers for the two tools to track interesting events. */
-extern void (*MAC_(new_mem_heap)) ( Addr a, UInt len, Bool is_inited );
-extern void (*MAC_(ban_mem_heap)) ( Addr a, UInt len );
-extern void (*MAC_(die_mem_heap)) ( Addr a, UInt len );
-extern void (*MAC_(copy_mem_heap))( Addr from, Addr to, UInt len );
+extern void (*MAC_(new_mem_heap)) ( Addr a, SizeT len, Bool is_inited );
+extern void (*MAC_(ban_mem_heap)) ( Addr a, SizeT len );
+extern void (*MAC_(die_mem_heap)) ( Addr a, SizeT len );
+extern void (*MAC_(copy_mem_heap))( Addr from, Addr to, SizeT len );
 
 /* Function pointers for internal sanity checking. */
-extern Bool (*MAC_(check_noaccess))( Addr a, UInt len, Addr* bad_addr );
+extern Bool (*MAC_(check_noaccess))( Addr a, SizeT len, Addr* bad_addr );
 
 /* Used in describe_addr() */
 extern Bool (*MAC_(describe_addr_supp))    ( Addr a, AddrInfo* ai );
 
 /* For VALGRIND_COUNT_LEAKS client request */
 extern Int MAC_(bytes_leaked);
+extern Int MAC_(bytes_indirect);
 extern Int MAC_(bytes_dubious);
 extern Int MAC_(bytes_reachable);
 extern Int MAC_(bytes_suppressed);
@@ -304,22 +330,22 @@ extern void MAC_(clear_MAC_Error)          ( MAC_Error* err_extra );
 
 extern Bool MAC_(shared_recognised_suppression) ( Char* name, Supp* su );
 
-extern void* MAC_(new_block) ( Addr p, UInt size, UInt align, UInt rzB,
+extern void* MAC_(new_block) ( Addr p, SizeT size, SizeT align, UInt rzB,
                                Bool is_zeroed, MAC_AllocKind kind,
                                VgHashTable table);
 extern void MAC_(handle_free) ( Addr p, UInt rzB, MAC_AllocKind kind );
 
 extern void MAC_(create_mempool)(Addr pool, UInt rzB, Bool is_zeroed);
 extern void MAC_(destroy_mempool)(Addr pool);
-extern void MAC_(mempool_alloc)(Addr pool, Addr addr, UInt size);
+extern void MAC_(mempool_alloc)(Addr pool, Addr addr, SizeT size);
 extern void MAC_(mempool_free)(Addr pool, Addr addr);
 
 extern void MAC_(record_address_error)     ( ThreadId tid, Addr a,
                                              Int size, Bool isWrite );
-extern void MAC_(record_core_mem_error)    ( ThreadId tid, Bool isWrite,
+extern void MAC_(record_core_mem_error)    ( ThreadId tid, Bool isUnaddr,
                                              Char* s );
-extern void MAC_(record_param_error)       ( ThreadId tid, Addr a,   
-                                             Bool isWriteLack, Char* msg );
+extern void MAC_(record_param_error)       ( ThreadId tid, Addr a, Bool isReg,
+                                             Bool isUnaddr, Char* msg );
 extern void MAC_(record_jump_error)        ( ThreadId tid, Addr a );
 extern void MAC_(record_free_error)        ( ThreadId tid, Addr a );
 extern void MAC_(record_freemismatch_error)( ThreadId tid, Addr a );
@@ -331,10 +357,10 @@ extern void MAC_(pp_shared_SkinError)      ( Error* err);
 extern MAC_Chunk* MAC_(first_matching_freed_MAC_Chunk)( Bool (*p)(MAC_Chunk*, void*), void* d );
 
 extern void MAC_(common_pre_clo_init) ( void );
-extern void MAC_(common_fini)         ( void (*leak_check)(void) );
+extern void MAC_(common_fini)         ( void (*leak_check)(LeakCheckMode mode) );
 
 extern Bool MAC_(handle_common_client_requests) ( ThreadId tid, 
-                                                  UInt* arg_block, UInt* ret );
+                                                  UWord* arg_block, UWord* ret );
 
 extern void MAC_(print_malloc_stats) ( void );
 
@@ -343,8 +369,9 @@ extern void MAC_(pp_LeakError)(void* vl, UInt n_this_record,
                                          UInt n_total_records); 
                            
 extern void MAC_(do_detect_memory_leaks) (
-          Bool is_valid_64k_chunk ( UInt ),
-          Bool is_valid_address   ( Addr )
+          LeakCheckMode mode,
+          Bool (*is_valid_64k_chunk) ( UInt ),
+          Bool (*is_valid_address)   ( Addr )
        );
 
 extern REGPARM(1) void MAC_(new_mem_stack_4)  ( Addr old_ESP );
@@ -357,8 +384,8 @@ extern REGPARM(1) void MAC_(new_mem_stack_16) ( Addr old_ESP );
 extern REGPARM(1) void MAC_(die_mem_stack_16) ( Addr old_ESP );
 extern REGPARM(1) void MAC_(new_mem_stack_32) ( Addr old_ESP );
 extern REGPARM(1) void MAC_(die_mem_stack_32) ( Addr old_ESP );
-extern            void MAC_(die_mem_stack) ( Addr a, UInt len);
-extern            void MAC_(new_mem_stack) ( Addr a, UInt len);
+extern            void MAC_(die_mem_stack) ( Addr a, SizeT len);
+extern            void MAC_(new_mem_stack) ( Addr a, SizeT len);
 
 
 /*------------------------------------------------------------*/
@@ -520,13 +547,13 @@ void REGPARM(1) MAC_(die_mem_stack_32)(Addr new_ESP)                          \
    }                                                                          \
 }                                                                             \
                                                                               \
-void MAC_(new_mem_stack) ( Addr a, UInt len )                                 \
+void MAC_(new_mem_stack) ( Addr a, SizeT len )                                \
 {                                                                             \
    PROF_EVENT(115);                                                           \
    UNALIGNED_NEW ( a, len );                                                  \
 }                                                                             \
                                                                               \
-void MAC_(die_mem_stack) ( Addr a, UInt len )                                 \
+void MAC_(die_mem_stack) ( Addr a, SizeT len )                                \
 {                                                                             \
    PROF_EVENT(125);                                                           \
    UNALIGNED_DIE ( a, len );                                                  \
